@@ -1,86 +1,35 @@
 """
 This script contains all function related to the implementation of the
 k-ToM agent.
-
-For reference the parameter order of the k-ToM parameters is:
-[1] Volatility
-[2] Behavioural temperature
-[3] Dilution
-[4] Bias
 """
+# For reference the parameter order of the k-ToM parameters is:
+# [1] Volatility
+# [2] Behavioural temperature
+# [3] Dilution
+# [4] Bias
+
+from typing import Tuple, Union
 import numpy as np
 from tomsup.payoffmatrix import PayoffMatrix
 import copy
-
-
-# Logit functions
-def inv_logit(x):
-    """
-    This is the inverse logit (or sigmoid) function from the original VBA
-    package
-    It is slightly different from a standard inverse logit. Most importantly,
-    it uses a finite precision, epsilon.
-    Note that the VBA implementation has more options8changing scale, offset,
-    etc.). These are not implemented here.
-    Note also that we have here used the "quick version" from the VBA package.
-    """
-    # Set precision parameter (0 means perfect precision)
-    epsilon = 1e-9
-
-    # Set input bounds
-    if np.any(x > 500):
-        # For scalars
-        if np.shape(x) == ():
-            x = 500
-        # Or vectors etc
-        else:
-            x[x > 500] = 500
-        # warn("Logit input constrained at upper bound 0.9999 to avoid
-        # rounding errors", Warning)
-    if np.any(x < -500):
-        if np.shape(x) == ():
-            x = -500
-        else:
-            x[x < -500] = -500
-        # warn("Logit input constrained at lower bound 0.0001 to avoid
-        # rounding errors", Warning)
-
-    # Calculate
-    y = epsilon + (1 - 2 * epsilon) / (1 + np.exp(-x))
-
-    return y
-
-
-def logit(x):
-    """
-    This is the the logit function from the original VBA package. See that
-    package for mor information.
-    """
-    # Calculate
-    lx = (x) ** -1 - 1
-    y = 0 - 1 ** -1 * np.log(lx)
-
-    return y
+from scipy.special import expit as inv_logit
+from scipy.special import logit
 
 
 # Learning subfunctions
-def p_op_var0_update(prev_p_op_mean0, prev_p_op_var0, volatility):
-    """
-    prev_p_op_mean0   (float)
-    prev_p_op_var0    (float)
-    volatility        (float)
 
-    0-ToM updates variance / uncertainty on choice probability estimate
+def p_op_var0_update(prev_p_op_mean0: float, prev_p_op_var0: float, volatility: float):
+    """Variance update of the 0-ToM
 
     Examples:
-    >>> p_op_var0_update(1, 0.2, 1)
-    0.8348496471878395
-    >>> #Higher volatility results in a higher variance
-    >>> p_op_var0_update(1, 0.2, 1) < p_op_var0_update (1, 0.2, 2)
-    True
-    >>> #Mean close to 0.5 gives lower variance
-    >>> p_op_var0_update(1, 0.45, 1) < p_op_var0_update (1, 0.2, 2)
-    True
+        >>> p_op_var0_update(1, 0.2, 1)
+        0.8348496471878395
+        >>> #Higher volatility results in a higher variance
+        >>> p_op_var0_update(1, 0.2, 1) < p_op_var0_update (1, 0.2, 2)
+        True
+        >>> #Mean close to 0.5 gives lower variance
+        >>> p_op_var0_update(1, 0.45, 1) < p_op_var0_update (1, 0.2, 2)
+        True
     """
     # Input variable transforms
     volatility = np.exp(volatility)
@@ -99,13 +48,8 @@ def p_op_var0_update(prev_p_op_mean0, prev_p_op_var0, volatility):
     return new_p_op_var0
 
 
-def p_op_mean0_update(prev_p_op_mean0, p_op_var0, op_choice):
-    """
-    prev_p_op_mean0 (float)
-    p_op_var0       (float)
-    op_choice       (int)
-
-    0-ToM updates mean choice probability estimate
+def p_op_mean0_update(prev_p_op_mean0: float, p_op_var0: float, op_choice: int):
+    """0-ToM updates mean choice probability estimate
     """
     # Input variable transforms
     p_op_var0 = np.exp(p_op_var0)
@@ -121,21 +65,20 @@ def p_op_mean0_update(prev_p_op_mean0, p_op_var0, op_choice):
     return new_p_op_mean0
 
 
-def p_opk_approx_fun(prev_p_op_mean, prev_param_var, prev_gradient, level):
-    """
-    prev_p_op_mean  (numpy.ndarray)
-    prev_param_var  (numpy.ndarray)
-    prev_gradient   (numpy.ndarray)
-    level           (int)
+def p_opk_approx_fun(
+    prev_p_op_mean: np.array,
+    prev_param_var: np.array,
+    prev_gradient: np.array,
+    level: int,
+):
 
+    """
     Approximates the estimated choice probability of the opponent on the
     previous round.
     A semi-analytical approximation derived in Daunizeau, J. (2017)
 
-    >>> p_opk_approx_fun(prev_p_op_mean = np.array([0]), \
-        prev_param_var = np.array([[0, 0, 0]]), \
-        prev_gradient = np.array([[0, 0, 0]]), level = 1)
-    array([-0.69314718])
+    Examples:
+        >>> p_opk_approx_fun(prev_p_op_mean = np.array([0]), prev_param_var = np.array([[0, 0, 0]]), prev_gradient = np.array([[0, 0, 0]]), level = 1)
     """
     # Constants
     a = 0.205
@@ -164,15 +107,16 @@ def p_opk_approx_fun(prev_p_op_mean, prev_param_var, prev_gradient, level):
     return p_opk_approx
 
 
-def p_k_udpate(prev_p_k, p_opk_approx, op_choice, dilution=None):
+def p_k_udpate(
+    prev_p_k: np.array, p_opk_approx: np.array, op_choice: int, dilution=None
+):
     """
     k-ToM updates its estimate of opponents sophistication level.
     If k-ToM has a dilution parameter, it does a partial forgetting of learned
     estimates.
 
-    >>> p_k_udpate(prev_p_k = np.array([1.]), \
-        p_opk_approx = np.array([-0.69314718]), op_choice = 1, dilution = None)
-    array([1.])
+    Examples:
+        >>> p_k_udpate(prev_p_k = np.array([1.]), p_opk_approx = np.array([-0.69314718]), op_choice = 1, dilution = None)
     """
     # Input variable transforms
     p_opk_approx = np.exp(p_opk_approx)
@@ -196,22 +140,24 @@ def p_k_udpate(prev_p_k, p_opk_approx, op_choice, dilution=None):
 
 
 def param_var_update(
-    prev_p_op_mean,
-    prev_param_var,
-    prev_gradient,
-    p_k,
-    volatility,
+    prev_p_op_mean: np.array,
+    prev_param_var: np.array,
+    prev_gradient: np.array,
+    p_k: np.array,
+    volatility: float,
     volatility_dummy=None,
     **kwargs
 ):
     """
     k-ToM updates its uncertainty / variance on its estimates of opponent's
     parameter values
-    >>> param_var_update(prev_param_mean = np.array([[0, 0, 0]]), \
-        prev_param_var = np.array([[0, 0, 0]]), \
-        prev_gradient = np.array([0, 0, 0]), p_k = np.array([1.]), \
-        volatility = -2, volatility_dummy = None)
-    array([[0.12692801, 0.        , 0.        ]])
+
+    Examples:
+        >>> param_var_update(prev_param_mean = np.array([[0, 0, 0]]), \
+            prev_param_var = np.array([[0, 0, 0]]), \
+            prev_gradient = np.array([0, 0, 0]), p_k = np.array([1.]), \
+            volatility = -2, volatility_dummy = None)
+        array([[0.12692801, 0.        , 0.        ]])
     """
     # Dummy constant: sets volatility to 0 for all except volatility opponent
     # parameter estimates
@@ -240,16 +186,18 @@ def param_var_update(
 
 
 def param_mean_update(
-    prev_p_op_mean, prev_param_mean, prev_gradient, p_k, param_var, op_choice
+    prev_p_op_mean: np.array,
+    prev_param_mean: np.array,
+    prev_gradient: np.array,
+    p_k: np.array,
+    param_var,
+    op_choice: int,
 ):
     """
     k-ToM updates its estimates of opponent's parameter values
 
-    >>> param_mean_update(prev_p_op_mean, \
-        prev_param_mean = np.array([[0, 0, 0]]), \
-        prev_gradient = np.array([0, 0, 0]), \
-        p_k = np.array([0, 0, 0]), param_var, op_choice)
-        array([[0.12692801, 0.        , 0.        ]])
+    Examples:
+        >>> param_mean_update(prev_p_op_mean, prev_param_mean = np.array([[0, 0, 0]]), prev_gradient = np.array([0, 0, 0]), p_k = np.array([0, 0, 0]), param_var, op_choice)
     """
     # Input variable transforms
     param_var = np.exp(param_var) * prev_gradient
@@ -280,12 +228,13 @@ def gradient_update(
     p_matrix,
     **kwargs
 ):
-    """"""
+    """The gradient update of the k-ToM agent
+    """
     # Make empty list for fillin in gradients
     gradient = np.zeros(len(param_mean))
 
     # The gradient is calculated for each parameter one at a time
-    for param in range(len(param_mean)):
+    for param, _ in enumerate(param_mean):
         # Calculate increment
         increment = max(abs(1e-4 * param_mean[param]), 1e-4)
         # Use normal parameter estimates
@@ -294,7 +243,8 @@ def gradient_update(
         param_mean_incr[param] = param_mean[param] + increment
 
         # Make parameter structure similar to own
-        sim_params_incr = copy.deepcopy(params)
+        # sim_params_incr = copy.deepcopy(params)
+        sim_params_incr = dict()
         # Populate it with estimated values, including the increment
         for param_idx, param_key in enumerate(params):
             sim_params_incr[param_key] = param_mean_incr[param_idx]
@@ -329,16 +279,14 @@ def gradient_update(
 
 
 # Decision subfunctions
-def p_op0_fun(p_op_mean0, p_op_var0):
+def p_op0_fun(p_op_mean0: float, p_op_var0: float):
     """
     0-ToM combines the mean and variance of its parameter estimate into a
     final choice probability estimate.
-    NB: This is the function taken from the VBA package (Daunizeau 2014),
-    which does not use 0-ToM's volatility parameter to avoid unidentifiability
-    problems.
+    To avoid unidentifiability problems this function does not use 0-ToM's volatility parameter.
 
+    Examples:
     >>> p_op0_fun(p_op_mean0 = 0.7, p_op_var0 = 0.3)
-    0.6397417553178626
     """
     # Constants
     a = 0.36
@@ -354,13 +302,11 @@ def p_op0_fun(p_op_mean0, p_op_var0):
     return p_op0
 
 
-def p_opk_fun(p_op_mean, param_var, gradient):
+def p_opk_fun(p_op_mean: np.array, param_var: np.array, gradient: np.array):
     """
     k-ToM combines the mean choice probability estimate and the variances of
     its parameter estimates into a final choice probability estimate.
-    NB: this is the function taken from the VBA package (Daunizeau 2014),
-    which does not use k-ToM's volatility parameter to avoid unidentifiability
-    issues.
+    To avoid unidentifiability problems this function does not use 0-ToM's volatility parameter.
     """
     # Constants
     a = 0.36
@@ -380,18 +326,22 @@ def p_opk_fun(p_op_mean, param_var, gradient):
     return p_opk
 
 
-def expected_payoff_fun(p_op, agent, p_matrix):
-    """
-    p_op (0 <= float <= 1): The probability of the opponent choosing 1
-    agent (0 <= int <= 1): the perspective of the agent
-    p_matrix (PayoffMatix): a payoff matrix
+def expected_payoff_fun(p_op: float, agent: int, p_matrix: PayoffMatrix):
+    """Calculate expected payoff of choosing 1 over 0
 
-    Calculate expected payoff of choosing 1 over 0
+
+    Args:
+        p_op (float): The probability of the opponent choosing 1
+        agent (int): The perspective of the agent
+        p_matrix (PayoffMatrix): A payoff matrix
+
+    Returns:
+        The expected payoff
 
     Examples:
-    >>> staghunt = PayoffMatrix(name = 'staghunt')
-    >>> expected_payoff_fun(1, agent = 0, p_matrix = staghunt)
-    2
+        >>> staghunt = PayoffMatrix(name = 'staghunt')
+        >>> expected_payoff_fun(1, agent = 0, p_matrix = staghunt)
+        2
     """
     # Calculate
     expected_payoff = p_op * (
@@ -401,7 +351,7 @@ def expected_payoff_fun(p_op, agent, p_matrix):
     return expected_payoff
 
 
-def softmax(expected_payoff, params):
+def softmax(expected_payoff, params: dict) -> float:
     """
     Softmax function for calculating own probability of choosing 1
     """
@@ -438,26 +388,28 @@ def softmax(expected_payoff, params):
 
 # Full learning and decision functions
 def learning_function(
-    prev_internal_states,
-    params,
-    self_choice,
-    op_choice,
-    level,
-    agent,
-    p_matrix,
+    prev_internal_states: dict,
+    params: dict,
+    self_choice: int,
+    op_choice: int,
+    level: int,
+    agent: int,
+    p_matrix: PayoffMatrix,
     **kwargs
-):
-    """
-    Examples:
-    >>> penny = PayoffMatrix(name = "penny_competitive")
-    >>> prev_internal_states = {'opponent_states': {}, \
-        'own_states': {'p_op_mean0': 0, 'p_op_var0': 0}}
-    >>> params = {'volatility': -2, 'b_temp': -1}
-    >>> learning_function(prev_internal_states, params, self_choice=1, \
-        op_choice=1, level=0, agent=0, p_matrix=penny)
-    {'opponent_states': {},
-    'own_states': {'p_op_mean0': 0.44216598162254866,
-    'p_op_var0': -0.12292276280308079}}
+) -> dict:
+    """The general learning function for the k-ToM agent
+
+    Args:
+        prev_internal_states (dict): Previous internal states
+        params (dict): The parameters
+        self_choice (int): Previous choice of the agent
+        op_choice (int): Opponents choice
+        level (int): sophistication level
+        agent (int): the perspective of the agent in the payoff matrix (0 or 1)
+        p_matrix (PayoffMatrix): a payoff matrix
+
+    Returns:
+        dict: The updated internal states
     """
     # Extract needed parameters
     volatility = params["volatility"]
@@ -528,7 +480,7 @@ def learning_function(
             )
 
             # Make parameter structure similar to own
-            sim_params = copy.deepcopy(params)
+            sim_params = dict()
 
             # Populate it with estimated values
             for param_idx, param_key in enumerate(params):
@@ -586,17 +538,34 @@ def learning_function(
     return new_internal_states
 
 
-def decision_function(new_internal_states, params, agent, level, p_matrix):
-    """
+def decision_function(
+    new_internal_states: dict,
+    params: dict,
+    agent: int,
+    level: int,
+    p_matrix: PayoffMatrix,
+) -> Tuple[float, float]:
+    """The decision function of the k-ToM agent
+
+    Args:
+        new_internal_states (dict): Dict of updated internal states
+        params (dict): The parameters
+        agent (int): the perspective of the agent in the payoff matrix
+        level (int): The sophistication level of the agent
+        p_matrix (PayoffMatrix): a payoff matrix
+
+    Returns:
+        Tuple[float, float]: a tuple contain probability of self choosing 1 and op choosing 1.
+
 
     Examples:
-    >>> penny = PayoffMatrix(name = "penny_competitive")
-    >>> new_internal_states = {'opponent_states': {}, \
-    'own_states': {'p_op_mean0': 30, 'p_op_var0': 2}}
-    >>> params = {'volatility': -2, 'b_temp': -1}
-    >>> decision_function(new_internal_states, params, agent = 0, \
-        level = 0, p_matrix = penny)
-    -5.436561973742046
+        >>> penny = PayoffMatrix(name = "penny_competitive")
+        >>> new_internal_states = {'opponent_states': {}, \
+        'own_states': {'p_op_mean0': 30, 'p_op_var0': 2}}
+        >>> params = {'volatility': -2, 'b_temp': -1}
+        >>> decision_function(new_internal_states, params, agent = 0, \
+            level = 0, p_matrix = penny)
+        -5.436561973742046
     """
     # If (simulated) agent is a 0-ToM
     if level == 0:
@@ -636,16 +605,30 @@ def decision_function(new_internal_states, params, agent, level, p_matrix):
 
 # Full k-ToM Function
 def k_tom(
-    prev_internal_states,
-    params,
-    self_choice,
-    op_choice,
-    level,
-    agent,
-    p_matrix,
+    prev_internal_states: dict,
+    params: dict,
+    self_choice: int,
+    op_choice: int,
+    level: int,
+    agent: int,
+    p_matrix: PayoffMatrix,
     **kwargs
-):
-    """"""
+) -> Tuple[int, dict]:
+    """The full k-ToM implementation
+
+    Args:
+        prev_internal_states (dict):  Dict of previous internal states
+        params (dict): The parameters
+        self_choice (int): the agent choice the previous round
+        op_choice (int): The opponents choice the previous round
+        level (int): The sophistication level of the agent
+        agent (int): the perspective of the agent in the payoff matrix
+        p_matrix (PayoffMatrix): a payoff matrix
+
+
+    Returns:
+        Tuple[int, dict]: a tuple containing the choice and the updated internal states
+    """
 
     # Update estimates of opponent based on behaviour
     if self_choice is not None:
@@ -682,10 +665,17 @@ def k_tom(
 
 
 # Initializing function
-def init_k_tom(params, level, priors="default"):
+def init_k_tom(params: dict, level: int, priors: Union[dict, str] = "default"):
     """
-    >>> init_k_tom(params = {'volatility': -2, 'b_temp': -1, 'bias':0 }, \
-        level=1, priors='default')
+    Initialization function of the k-ToM agent
+
+    Args:
+        params (dict): The starting parameters
+        level (int): The sophistication level of the agent
+        priors (Union[dict, str], optional): The priors of the k-ToM. Default to "default". 
+            See tutorial on how to set internal states of the k-ToM agent.
+    Examples:
+        >>> init_k_tom(params = {'volatility': -2, 'b_temp': -1, 'bias':0 }, level=1, priors='default')
     """
     # If no priors are specified
     if priors == "default":
@@ -746,39 +736,3 @@ def init_k_tom(params, level, priors="default"):
     internal_states["own_states"] = own_states
 
     return internal_states
-
-
-# Testing function
-if __name__ == "__main__":
-    import doctest
-
-    doctest.testmod(verbose=True)
-
-
-if __name__ == "__main__":
-    p_matrix = PayoffMatrix(name="penny_competitive")
-
-    prev_internal_states = {
-        "opponent_states": {
-            0: {"opponent_states": {}, "own_states": {"p_op_mean0": 0, "p_op_var0": 0}}
-        },
-        "own_states": {
-            "p_k": np.array([1.0]),
-            "p_op_mean": np.array([0]),
-            "param_mean": np.array([[0, 0, 0]]),
-            "param_var": np.array([[0, 0, 0]]),
-            "gradient": np.array([[0, 0, 0]]),
-        },
-    }
-
-    params = {"volatility": -2, "b_temp": -1, "bias": 0}
-
-    k_tom(
-        prev_internal_states,
-        params,
-        self_choice=1,
-        op_choice=1,
-        level=1,
-        agent=0,
-        p_matrix=p_matrix,
-    )
